from enum import IntEnum


class Region(IntEnum):
    pass


class SnowRegion(Region):
    SVALBARD_OST = 3001
    SVALBARD_VEST = 3002
    NORDENSKIOLD_LAND = 3003
    SVALBARD_SOR = 3004
    OST_FINNMARK = 3005
    FINNMARKSKYSTEN = 3006
    VEST_FINNMARK = 3007,
    FINNMARKSVIDDA = 3008
    NORD_TROMS = 3009
    LYNGEN = 3010
    TROMSO = 3011
    SOR_TROMS = 3012
    INDRE_TROMS = 3013
    LOFOTEN_OG_VESTERALEN = 3014
    OFOTEN = 3015
    SALTEN = 3016
    SVARTISEN = 3017
    HELGELAND = 3018
    NORD_TRONDELAG = 3019
    SOR_TRONDELAG = 3020
    YTRE_NORDMORE = 3021
    TROLLHEIMEN = 3022
    ROMSDAL = 3023
    SUNNMORE = 3024
    NORD_GUDBRANDSDALEN = 3025
    YTRE_FJORDANE = 3026
    INDRE_FJORDANE = 3027
    JOTUNHEIMEN = 3028
    INDRE_SOGN = 3029
    VOSS = 3031
    HALLINGDAL = 3032
    HORDALANDSKYSTEN = 3033
    HARDANGER = 3034
    VEST_TELEMARK = 3035
    ROGALANDSKYSTEN = 3036
    HEIANE = 3037
    AGDER_SOR = 3038
    TELEMARKS_SOR = 3039
    VESTFOLD = 3040
    BUSKERUD_SOR = 3041
    OPPLAND_SOR = 3042
    HEDMARK = 3043
    AKERSHUS = 3044
    OSLO = 3045
    OSTFOLD = 3046


